
#ifndef __soramimi_jp__buffer_h
#define __soramimi_jp__buffer_h

namespace text {
	template <typename T> T const *zerostring();
	template <> inline char const *zerostring<char>() { return ""; }
	template <> inline wchar_t const *zerostring<wchar_t>() { return L""; }

	template <typename T> size_t length(T const *p);
	template <> inline size_t length<char>(char const *p) { return strlen(p); }
	template <> inline size_t length<wchar_t>(wchar_t const *p) { return wcslen(p); }
}

template <typename T> class string_buffer {
private:
	struct fragment_t {
		fragment_t *next;
		size_t size;
		size_t used;
		T data[1];
	};
	mutable fragment_t *_fragment;
private:
	string_buffer(string_buffer const &r);
	void operator = (string_buffer const &r);

	static void store(T const *ptr, T const *end, T *dst)
	{
		while (ptr < end) {
			new (dst) T(*ptr);
			ptr++;
			dst++;
		}
	}
	void internal_clear() const
	{
		while (_fragment) {
			fragment_t *next = _fragment->next;
			T *ptr = _fragment->data;
			T *end = _fragment->data + _fragment->used;
			while (ptr < end) {
				ptr->~T();
				ptr++;
			}
			delete (unsigned char *)_fragment;
			_fragment = next;
		}
	}
	T *internal_get() const
	{
		if (!_fragment) {
			return 0;
		}
		if (_fragment->next) {
			size_t len = size();
			fragment_t *newptr = (fragment_t *)new unsigned char [sizeof(fragment_t) + sizeof(T) * len];
			newptr->next = 0;
			newptr->size = len;
			newptr->used = len;
			memset(&newptr->data[len], 0, sizeof(T));
			fragment_t *f = _fragment;
			while (f && len > 0) {
				len -= f->used;
				store(f->data, f->data + f->used, newptr->data + len);
				f = f->next;
			}
			internal_clear();
			_fragment = newptr;
		}
		return _fragment->data;
	}
public:
	string_buffer()
		: _fragment(0)
	{
	}
public:
	~string_buffer()
	{
		clear();
	}
	void clear()
	{
		internal_clear();
	}
	void write(T const *ptr, size_t len)
	{
		if (ptr) {
			if (len > 0) {
				if (_fragment && _fragment->size > _fragment->used) {
					size_t n = _fragment->size - _fragment->used;
					if (n > len) {
						n = len;
					}
					store(ptr, ptr + n, _fragment->data + _fragment->used);
					_fragment->used += n;
					_fragment->data[_fragment->used] = 0;
					ptr += n;
					len -= n;
				}
				if (len > 0) {
					size_t n = 4096;
					if (n < len) {
						n = len;
					}
					fragment_t *newptr = (fragment_t *)new unsigned char [sizeof(fragment_t) + sizeof(T) * n];
					newptr->next = _fragment;
					newptr->size = n;
					newptr->used = len;
					store(ptr, ptr + len, newptr->data);
					newptr->data[newptr->used] = 0;
					_fragment = newptr;
				}
			}
		}
	}
	void write(T const *ptr)
	{
		write(ptr, text::length(ptr));
	}
	void write(T const &t)
	{
		write(&t, 1);
	}
	void write(std::basic_string<T> const &str)
	{
		write(str.c_str(), str.size());
	}
	size_t size() const
	{
		size_t len = 0;
		fragment_t *f = _fragment;
		while (f) {
			len += f->used;
			f = f->next;
		}
		return len;
	}
	T const *get() const
	{
		T *p = internal_get();
		return p ? p : text::zerostring<T>();
	}
	bool empty() const
	{
		return size() == 0;
	}
	std::basic_string<T> str() const
	{
		T const *p = get();
		size_t n = size();
		return std::basic_string<T>(p, n);
	}
};


#endif

