
#ifndef __httpserver_h
#define __httpserver_h

#include <map>
#include <string>
#include <vector>

#include "thread.h"
#include "event.h"
#include "Socket.h"
#include "httpstatus.h"

namespace httpserver {

	struct http_request_t {
		std::vector<std::string> header;
	};

	struct http_response_t {
		std::vector<std::string> header;
		std::vector<unsigned char> buffer;
		void set_content_type(std::string const &ct);
		http_status_t const *set_location(std::string const &url);
	};


	typedef http_status_t const *(*http_get_handler_t)(void *cookie, std::string const &uri, http_request_t *request, http_response_t *response);

	struct http_get_handler_item_t {
		http_get_handler_t handler;
		void *cookie;
		http_get_handler_item_t()
			: handler(0)
			, cookie(0)
		{
		}
		http_get_handler_item_t(http_get_handler_t handler, void *cookie)
			: handler(handler)
			, cookie(cookie)
		{
		}
	};

	struct SuffixHandler {
		virtual http_status_t const *process(std::string const &uri, std::string const &suffix, http_response_t *response) = 0;
	};

	struct HTTP_Server : public Thread {
		int tcp_port;
		std::string document_root;
		Socket listener_socket;

		Event listener_event;

		HTTP_Server();

		std::vector<http_get_handler_item_t> http_get_handler_table;

		http_status_t const *HTTP_Server::dispatch_get_request(std::string const &uri, http_request_t *request, http_response_t *response);

		void add_http_get_handler(http_get_handler_t method, void *cookie)
		{
			http_get_handler_table.push_back(http_get_handler_item_t(method, cookie));
		}

		void set_tcp_port(int port)
		{
			tcp_port = port;
		}

		void set_document_root(std::string const &docroot)
		{
			document_root = docroot;
		}

		class RequestHandlerMap {
		private:
			std::map<std::string, SuffixHandler *> map;
		public:
			RequestHandlerMap()
			{
			}
			void add(std::string const &suffix, SuffixHandler *handler)
			{
				map[suffix] = handler;
			}
			SuffixHandler *find(std::string const &suffix) const
			{
				std::map<std::string, SuffixHandler *>::const_iterator it = map.find(suffix);
				if (it == map.end()) {
					return 0;
				}
				return it->second;
			}
		};

		RequestHandlerMap suffix_handler_map;

		class ContentTypeMap {
		private:
			std::map<std::string, std::string> map;
		public:
			ContentTypeMap()
			{
				add("txt", "text/plain");
				add("html", "text/html");
				add("html", "text/html");
				add("css", "text/css");
				add("jpg", "image/jpeg");
				add("jpeg", "image/jpeg");
				add("png", "image/png");
				add("gif", "image/gif");
			}
			void add(std::string const &suffix, std::string const &mimetype)
			{
				map[suffix] = mimetype;
			}
			std::string find(std::string const &suffix) const
			{
				std::map<std::string, std::string>::const_iterator it = map.find(suffix);
				if (it == map.end()) {
					return std::string();
				}
				return it->second;
			}
		};

		ContentTypeMap content_type_map;

		http_status_t const *default_handler(std::string const &uri, std::string const &suffix, http_response_t *response);
		http_status_t const *http_process_request(http_request_t *request, http_response_t *response);
		void http_send_response_header(Socket sock, http_status_t const *status, http_response_t *response);

		void run();
		virtual void stop();
	};
}

#endif

